/*
 * bounce.h:	Fast ring buffer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2002,2003 Werner Fink, <werner@suse.de>
 */

#ifndef __BOUNCE_H
#define __BOUNCE_H

#include <string.h>
#include <vdr/thread.h>
#include "types.h"

class cBounce {
// Special ring buffer with memmove on `get' side
// which allows to append only on `put' side
private:
    uint_8 *data;
    uint_8 *head;
    uint_8 *tail;
    const uint_8 *const end;
    volatile ssize_t avail;
    volatile ssize_t threshold;
    cMutex   ctr, io;
    cCondVar iowatch;
    bool sleeping;
    inline void reset(void) { tail = head = data; avail = 0; sleeping = false; };
public:
    cBounce(uint_8 *buf, const size_t len) : data(buf), end(buf+len) { reset(); };
    inline bool store(const uint_8 *const buf, const ssize_t len, const bool wakeup = true)
    {
	bool ok = true;
        ctr.Lock();
	if (len > end - tail) {
	    reset();
	    dsyslog("BOUNCE: buffer overflow\n");
	    debug("BOUNCE: buffer overflow %d > %d\n", len, end-tail);
	    ok = false;
	    goto out;
	}
	memcpy(tail, buf, len);
	avail = (tail += len) - head;
    out:
	if (wakeup && sleeping) iowatch.Broadcast();
	ctr.Unlock();
	return ok;
    }
    inline ssize_t fetch(uint_8 *buf, const size_t len)
    {
	ssize_t ret;
        ctr.Lock();
	if (!(ret = avail))
	    goto out;
	if ((size_t)ret > len)
	    ret = len;
	memcpy(buf, head, ret);
	head += ret;
	//
	// Ringbuffer arithmetics, even just for the case if fetch()
	// would take the length of *buf with length < avail
	//
	if (head >= tail) {	// empty, reset buffer
	    reset();
	    goto out;
	}
	if (head > data) {	// buffer not empty, move contents
	    avail = tail - head;
	    head = (uint_8 *)memmove(data, head, avail);
	    tail = head + avail;
	}
    out:
	ctr.Unlock();
	return ret;
    }
    const inline void flush(void) { ctr.Lock(); reset(); ctr.Unlock(); };
    const inline void bank(ssize_t val) { threshold = val; };
    const inline void signal(void) { /* if (!io) */ iowatch.Broadcast(); };
    const inline bool poll(int msec)
    {
	bool ret = true;
	if (avail <= threshold) {
	    sleeping = true;
	    ret = iowatch.TimedWait(io, msec);
	    sleeping = false;
	} else
	    pthread_yield();
	return ret;
    }
    const inline void takeio (void) { io.Lock();   flush(); };
    const inline void leaveio(void) { flush(); io.Unlock(); };
    const inline bool free(const size_t min) const {  return (size_t)(end - tail) > min; }
    const inline size_t getfree() const {  return (size_t)(end - tail); }
    const inline size_t getused() const {  return (size_t)(tail- data); }
};
#endif // __BOUNCE_H
