/*
 * replay.c:	Forward AC3 stream to the S/P-DIF interface of
 *		sound cards supported by ALSA. (NO decoding!)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2002,2003 Werner Fink, <werner@suse.de>
 *
 * Changes and Additional Code:
 *
 * Copyright (C) 2003 Sven Goethel, <sven@jausoft.com>
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include "types.h"
#include "replay.h"
#include "channel.h"
#include "ac3.h"
#include "dts.h"
#include "lpcm.h"

// --- cReplayOutAC3 : Forward AC3 stream to S/P-DIF of a sound card with ALSA -----------

const uint_32 cReplayOutSPDif::PS1magic = char2int(0x00,0x00,0x01,0xBD);
const uint_16 cReplayOutSPDif::AC3magic = char2short(0x0b,0x77);
const uint_32 cReplayOutSPDif::DTSmagic = char2int(0x7f,0xfe,0x80,0x01);
uint_8  * cReplayOutSPDif::pesdata;
cBounce * cReplayOutSPDif::bounce;

cReplayOutSPDif::cReplayOutSPDif(spdif &dev, ctrl_t &up, cBounce * bPtr, const char *script)
: SPDIFmute(script), spdifDev(&dev), setup(up)
{
    flags = 0;
    stream = NULL;
    bounce = bPtr;
    bounce->bank(0);
    pesdata = setup.buf + TRANSFER_START;
    PrimaryDevice = cDevice::PrimaryDevice();
}

cReplayOutSPDif::~cReplayOutSPDif(void)
{
    debug("cReplayOutSPDif::~ called\n");
    Clear();
}

void cReplayOutSPDif::Activate(bool onoff)
{
    if (onoff) {
        debug("cReplayOutSPDif::Activate 1\n");
	if (*spdifDev) {
	    debug("Activated failed due S/P-DIF already open\n");
	    goto out;		// An other has opened the S/P-DIF
	}
	if (test_setup(RESET)) {
	    clear_flag(FAILED);
	    clear_setup(RESET);
	}
	if (test_flag(FAILED) || test_flag(ACTIVE)) {
	    debug("Activated failed due set %s\n", test_flag(FAILED) ? "FAILED" : "ACTIVE");
	    goto out;
	}
	set_flag(ACTIVE);
	Start();
	pthread_yield();
        debug("cReplayOutSPDif::Activate X\n");
    } else {
        debug("cReplayOutSPDif::De-Activate 1\n");
	clear_flag(ACTIVE);
	while (test_flag(RUNNING)) {
            debug("cReplayOutSPDif::De-Activate 3\n");
	    clear_flag(ACTIVE);
	    bounce->flush();
	    bounce->signal();
	    usleep(1000);
	    if ( ! Active() ) break;  // S.O.S
	}
	Cancel(1);
	stream = NULL;
        debug("cReplayOutSPDif::De-Activate X\n");
    }
out:
    return;
}

void cReplayOutSPDif::Action(void)
{
    debug("cReplayOutSPDif::Action 1\n");
    set_flag(RUNNING);
    iec60958* tmp;
    if (!stream || !spdifDev->Open(stream, setup)) {
	esyslog("REPLAY: can't open spdif interface");
	debug("REPLAY: Action start failed: stream=%d\n", stream!=NULL);
	set_flag(FAILED);
	goto out;
    }

    dsyslog("REPLAY: Forwarding bitstream thread started (pid=%d)", getpid());
    realtime("INSTREAM: ");
    bounce->takeio();
    pts.clear();
    tmp = stream;
    while (test_flag(ACTIVE)) {
	ssize_t len;

	if (test_setup(MUTE)) {
	    pts.clear();
	    spdifDev->Clear();
	}

	if (!spdifDev->Synchronize(bounce))	// Wait on spdifDev->Signal()
	    break;

	if (test_setup(MUTE)) {
	    bounce->flush();
	    clear_flag(BOUNDARY);
	    continue;
	}

	if ((len = bounce->fetch(pesdata, spdifDev->Available(TRANSFER_MEM)))) {
	    /* int count = */ spdifDev->Forward(pesdata, len, bounce);
	}

    }
    spdifDev->Close();
    dsyslog("REPLAY: Forwarding bitstream thread ended (pid=%d)", getpid());
    tmp->Reset();
out:
    stream = NULL;
    bounce->leaveio();
    clear_flag(RUNNING);
}

inline bool cReplayOutSPDif::OffsetToPayOfPS1(const uchar *const b, int &off, const int cnt)
{
    bool ret = true;				// DVB streams do not have sub stream headers
    const uchar *const p = &b[off];

    // Should we check the stream number (p[0] & 0x07)?
    // Normally this plugin gets only one audio stream.

    switch ((p[0] & 0xf8)) {
    case 0x80:					// AC3 from e.g. DVD
	if (stream != &ac3 || stream->track != (p[0] & 7)+1)
	    ret = false;
	off += 4;
	break;
    case 0x88:					// DTS
	if (stream != &dts || stream->track != (p[0] & 7)+1)
	    ret = false;
	off += 4;
	break;
    case 0xA0:					// lPCM
	if (stream != &pcm) {
	    uint_32 ui = (uint_32)(&p[1]);
	    if (ui == 0 && p[5] == 0) {
		off = cnt;
		goto dvd;			// Don't support broken streams (dvd plugin).
	    }
	}
	if (stream != &pcm || stream->track != (p[0] & 7)+1)
	    ret = false;
	off += 7;
	break;
    default:					// Should not happen, e.g. stream
	ret = false;				// is broken. Should we reset???
	break;
    }

    if (cnt < off+1)
	ret = false;
dvd:
    return ret;
}

inline bool cReplayOutSPDif::ScanPayOfPS1(const uchar *const b, int &off, const int cnt)
{
    const uchar *const p = &b[off];
    const uchar *m;

    if (stream)
    {
    	debug("ScanPayOfPS1 stream .. -> out\n");
	goto out;
    }

#define available(x)	if (cnt < off+(x)) goto out

    // Should we check the stream number (p[0] & 0x07)?
    // Normally this plugin gets only one audio stream.

    switch ((p[0] & 0xf8)) {
    case 0x80:					// AC3 from e.g. DVD
	available(4);
	off += (((uint_16)p[2]<<8)|p[3]) + 3;
	available(2);

	m = &b[off];
	if (char2short(m[0],m[1]) == AC3magic) {
	    stream = &ac3;
	    stream->stream_type = IEC_AC3;
	    stream->Reset();
	    stream->isDVD = true;
	    stream->track = (p[0] & 7)+1;
	    debug("cReplayOutSPDif::ScanPayOfPS1 AC3 set atrack 0x%X -> %d\n",
			(int)p[0], stream->track);
	    clear_setup(AUDIO);
	}

	break;
    case 0x88:					// DTS
	available(4);
	off += (((uint_16)p[2]<<8)|p[3]) + 3;
	available(4);

	m = &b[off];
	if (char2int(m[0],m[1],m[2],m[3])  == DTSmagic) {
	    stream = &dts;
	    stream->stream_type = IEC_DTS;
	    stream->Reset();
	    stream->isDVD = true;
	    stream->track = (p[0] & 7)+1;
	    debug("cReplayOutSPDif::ScanPayOfPS1 DTS set atrack 0x%X -> %d\n",
			(int)p[0], stream->track);
	    clear_setup(AUDIO);
	}

	break;
    case 0xA0:					// lPCM
	debug("cReplayOutSPDif::ScanPayOfPS1 try lPCM len=%d.. ", cnt);

	available(8);
	if ((((uint_16)p[2]<<8)|p[3]) > 3)	// Offset to first channel 0 (left) sample
	    off += (((uint_16)p[2]<<8)|p[3]) + 3;
	else {
	    uint_32 ui = (uint_32)(&p[1]);
	    if (ui == 0 && p[5] == 0)
		break;				// Don't support broken streams (dvd plugin).
	    off += 7;
	}
	available(1);

	stream = &pcm;
        stream->stream_type = IEC_PCM;
	stream->Reset();
	stream->isDVD = true;
        stream->track = (p[0] & 7)+1;
	debug("cReplayOutSPDif::ScanPayOfPS1 PCM set atrack 0x%X -> %d\n",
			(int)p[0], stream->track);

	switch (p[5] & 0x30) {
	case 0x00: pcm.sample_rate = 48000;
	    break;
	case 0x20: pcm.sample_rate = 44100;
	    break;
	case 0x30: pcm.sample_rate = 32000;
	    break;
	case 0x10:				// 96kHz currently not supported
	default:   stream = NULL;
	    break;
	}
	switch (p[5] & 0xC0) {
	case 0x00:				// 16 bit samples
	    break;
	case 0x40:				// 20 bit currently not supported
	case 0x80:				// 24 bit currently not supported
	default:   stream = NULL;
	    break;
	}
	switch (p[5] & 0x07) {
	case 0x00:				// 1 channel
	case 0x01:				// 2 channels
	    break;
	default:   stream = NULL;		// More channels not supported
	    break;
	}

	if (!stream)
	    goto out;
	set_setup(AUDIO);

	break;
    default:					// e.g. AC3 emmbedded within DVB PS1
	available(2);
	if (char2short(p[0],p[1]) == AC3magic) {
	    stream = &ac3;
	    stream->stream_type = IEC_AC3;
	    stream->Reset();
	    stream->isDVD = false;
	    stream->track = (p[0] & 7)+1;
	    debug("cReplayOutSPDif::ScanPayOfPS1 emb. AC3 set atrack 0x%X -> %d\n",
			(int)p[0], stream->track);

	    clear_setup(AUDIO);
	}
	break;
    }

#undef available

    goto okout;

out:
    debug("invalid\n");
okout:
    return (stream != NULL);
}

inline int cReplayOutSPDif::DigestPayOfMP2(const uchar *const pstart, const int cnt)
{
        static unsigned long Frame = 0;
	uchar ptsflag = 0;

	debug("cReplayOutSPDif::DigestPayOfMP2 got PS-MP2-magic audio 0x%X..\n",
		(int)(pstart[3]));

        int datalen;
        const uint_8 * data = pstart;

	if (!test_flag(BOUNDARY))
	{
		debug("cReplayOutSPDif::DigestPayOfMP2 ! BOUNDARY data reset\n");
		MP2Dec.init();
	} else if ( stream->track != (pstart[3] & 7)+1 ) {
		debug("cReplayOutSPDif::DigestPayOfMP2 MP2 wrong atrack got %d(0x%X), locked %d\n",
			(int)(pstart[3] & 7)+1, (int)pstart[3], stream->track);
		goto done; // 0 skip
        }

	ptsflag = pstart[7] >> 6;
	if( ptsflag ) {
		int64_t devSTC;
		if ( (devSTC=cDevice::PrimaryDevice()->GetSTC())>=0 )
		        pts.process(&pstart[9], (uint_32)(devSTC/90UL), 0);
/**
		        if ( pts.process(&pstart[9], (uint_32)(devSTC/90UL), 0) > 5 )
			     	goto done;
 */
	}

	datalen  = (int)( (pstart[4] << 8) + pstart[5] + 6 ); // cPStream::packetLength

	datalen -= (int)  (pstart[8]);	// cPStream::PESHeaderLength
	data    += (int)  (pstart[8]); // cPStream::PESHeaderLength

	datalen -= 9; // //skip mandatory header bytes
	data    += 9; // //skip mandatory header bytes

	Frame++;

	MP2Dec.pushData ( data, datalen );

	while ( MP2Dec.decode() )
	{
	    if (!test_flag(BOUNDARY) )
	    {
		set_flag(BOUNDARY);

	        if ( stream == NULL )
		{
			stream = &pcm;
			stream->stream_type = IEC_PCMCVRTED_MP2;
			stream->Reset();
			stream->isDVD = true;
			stream->track = (pstart[3] & 7)+1 ;

			pcm.sample_rate = MP2Dec.sample_rate;

			debug("cReplayOutSPDif::DigestPayOfMP2 MP2 set atrack 0x%X -> %d\n",
				(int)pstart[3], stream->track);

			set_setup(AUDIO);
		} else {
			debug("cReplayOutSPDif::DigestPayOfMP2 !BOUNDARY stream allready set\n");
		}

		if (!test_flag(RUNNING))
			Activate(true);
	    }

	    if( MP2Dec.pcm_netto_size > 0 ) {
		    if (!bounce->store((uint_8 *)(MP2Dec.lpcmFrame().Data), 
				       MP2Dec.pcm_netto_size,
					bounce->getused()==0))
		    {
			goto skip; // skip 
		    }
	    }
	}
	goto done;
done:
	return 0; // 0 skip
skip:
        return 2;
}

void cReplayOutSPDif::Play(const uchar *b, int cnt)
{
    static uint_16 skip = 0;
    int len ;

    if (skip) {
	debug("cReplayOutSPDif::Play(skip=%d)\n", skip);
	skip--;
	goto out;
    }

    if (test_setup(MUTE) || test_setup(LIVE))
	goto out;

    if (!test_setup(ACTIVE)) {
	Clear();
	goto out;
    }

    if (cnt < 10)
	goto out;

    if ((b[6] & 0xC0) != 0x80)
    {
	debug("cReplayOutSPDif::Play 0x%X / 0x%X not 0x80 (PS1) \n", 
		(int)(b[6]), (int)(b[6]&0xC0));
	goto out;			// Not PS1 Mpeg2 PES
    }

    if (char2int(b[0],b[1],b[2],b[3]) == PS1magic)
    {
	len = (((uint_16)b[4]<<8)|b[5]) + 6;
	int off = b[8] + 9;

	if (cnt < off+1)
	    goto out;

        if (stream!=NULL && stream->stream_type >= IEC_PCMCVRTED) {
                debug("REPLAY: got PS1 (atrack 0x%X, 0x%X) allready choosen cvrt -> skip\n",
         	    b[3], b[off]);
	        goto out;
	}

	if (!test_flag(BOUNDARY)) {

	    // If the stream provides PTS/DTS information and/or
	    // the data in the pay load is marked as aligned we
	    // assume that that we'll find immediately the sync words
	    // of the pay load its self or the corresponding sub
	    // stream identification with the offset to the pay load.
	    
	    if ((b[7] & 0xC0) || (b[6] & 0x04)) {

		if (!ScanPayOfPS1(&b[0], off, cnt))
		    goto out;

		set_flag(BOUNDARY);	// At the beginning of a stream

		if (!test_flag(RUNNING))
		    Activate(true);
#if 0
		if (b[7] & 0xC0) {
		    // Do we need a sanity check? If so we have to
		    // use a global variable count (locking!)
		    // for increasing it here and decreasing it
		    // at Action()
		    XXXX
	        }
#endif
	    } else
		goto out;		// Start only at a boundary of a stream

	} else if (stream && stream->isDVD) {

	    if (!OffsetToPayOfPS1(b, off, cnt)) {
		Clear();		// Reset
		goto out;
	    }

	    if (off == cnt)
		goto out;		// Don't support broken streams (dvd plugin).
	}

	if (!bounce->store((uint_8 *)&b[off], min(len - off, cnt), true))
	    skip = 2;
    } 
    else if ( test_setup(MP2ENABLE) &&
              b[0]==0x00 && b[1]==0x00 && b[2]==0x01 && 
	      0xC0 <= b[3] && b[3] <= 0xDF
            ) 
    {
        if (stream!=NULL && stream->stream_type < IEC_PCMCVRTED) {
                debug("REPLAY: got PS-MP2 (atrack 0x%X) allready choosen cvrt -> skip\n", b[3]);
	        goto out;
	}

    	skip = DigestPayOfMP2(b, cnt);
	debug("cReplayOutSPDif::Play MP2 digested, skip=%d\n", skip);
    } 
    else
	debug("cReplayOutSPDif::Play NO Ps1magic ..\n");
out:
    return;
}

void cReplayOutSPDif::Mute(bool onoff)
{
    debug("cReplayOutSPDif::Mute(%d) called\n", onoff);
    if ((test_setup(MUTE) ? true : false) == onoff)
	goto out;

    if (onoff) {
	LOCK_THREAD;
	set_setup(MUTE);
	spdifDev->Pause(onoff);
	bounce->flush();
	bounce->signal();
	clear_flag(BOUNDARY);
    } else
	clear_setup(MUTE);

    if (SPDIFmute) {	// Normal TV mode
	char *cmd = NULL;
	asprintf(&cmd, "%s %s 2> /dev/null", SPDIFmute, test_setup(MUTE) ? "mute" : "unmute");
	if (cmd) {
	    system(cmd);
	    free(cmd);
	    cmd = NULL;
	}
    }
out:
    return;
}

void cReplayOutSPDif::Clear(void)
{
    LOCK_THREAD;
    debug("cReplayOutSPDif::Clear() called\n");
    (void)Activate(false);
    bounce->flush();
    bounce->signal();
    MP2Dec.init();
    clear_flag(BOUNDARY);
    Mute(cDevice::PrimaryDevice()->IsMute());
}
